#include "common.hpp"

bool pairCompare(const pair<int, double>& firstElem, const pair<int, double>& secondElem) 
{
	return firstElem.second > secondElem.second;
}

FILE* fopen_(const char* p, const char* m)
{
	FILE* f = fopen(p, m);
	if (!f) {
		printf("Failed to open %s\n", p);
		exit(1);
	}
	return f;
}

void kmeans(double** data, int N, int dim, int K, double thre, double**& centroids, int*& labels)
{
	labels = new int [N]; // label of each datapoint
	centroids = new double* [K]; // centroids
	double** centroids1 = new double* [K]; // temp centroids
	int* counts = new int [K]; // size of each cluster

	double prev_sum_of_dist, sum_of_dist = DBL_MAX; // sum of squared Euclidean distance

	// initialization
	for (int k = 0; k < K; k ++) {
		centroids[k]  = new double [dim];
		centroids1[k] = new double [dim];

		/// here pick a point as one initial centroids
		int which_pt = rand() % N;
		for (int d = 0; d < dim; d ++) {
			centroids[k][d] = data[which_pt][d];
		}
	}

	// main loop
	do {
		// save sum_of_dist from last step
		prev_sum_of_dist = sum_of_dist;
		sum_of_dist = 0;

		// clear old counts and temp centroids
		for (int k = 0; k < K; k ++) {
			for (int d = 0; d < dim; d ++) {
				centroids1[k][d] = 0;
			}
			counts[k] = 0;
		}

		// each point re-finds the closest cluster
		for (int i = 0; i < N; i ++) {
			double min_dist = DBL_MAX;
			for (int k = 0; k < K; k ++) {
				// calculate Euclidean distance
				double dist = 0;
				for (int d = 0; d < dim; d ++) {
					dist += square(data[i][d] - centroids[k][d]);
				}
				if (dist < min_dist) {
					labels[i] = k;
					min_dist = dist;
				}
			}
			// update size and temp centroids of the destination cluster
			for (int d = 0; d < dim; d ++) {
				centroids1[labels[i]][d] += data[i][d];
			}
			counts[labels[i]] ++;
			// update sum_of_dist
			sum_of_dist += min_dist;
		}

		// update all centroids
		for (int k = 0; k < K; k ++) {
			for (int d = 0; d < dim; d ++) {
				centroids[k][d] = counts[k] > 0 ? centroids1[k][d] / counts[k] : centroids1[k][d];
			}
		}

		// print progress
		fprintf(stderr, "X");
		fflush(stderr);
	} while (fabs(sum_of_dist - prev_sum_of_dist) > thre);

	for (int k = 0; k < K; k ++) {
		delete [] centroids1[k];
	}
	delete [] centroids1;
	delete [] counts;
}
