#include "corpus.hpp"

// rank the clicks in terms of timestamps
bool timeCompare(const pair<int,int>& firstElem, const pair<int,int>& secondElem) 
{
	return firstElem.second < secondElem.second;
}

void corpus::loadData(const char* clickFile, const char* imgFeatPath, const char* ownerPath, int userMin)
{
	nItems = 0;
	nUsers = 0;
	nClicks = 0;

	imFeatureDim = 4096;

	loadOwner(ownerPath);
	loadClicks(clickFile, userMin);	// determine I and U at this step
	loadImgFeatures(imgFeatPath); // finally load the features for only those valid items

	printf("\n  \"nUsers\": %d, \"nItems\": %d, \"nClicks\": %d\n", nUsers, nItems, nClicks);

	// statsitics();
}

void corpus::statsitics()
{
	int timeout = 3600; // 1h
	long long total_sessions = 0;
	long long total_sessions_len = 0;
	long long total_gaps = 0;
	long long total_gaps_len = 0;
	long long total_sessions_clicks = 0;
	long long total_transitions = 0;
	long long total_user_transitions = 0;
	for (int u = 0; u < nUsers; u ++) {
		long long session = 0;
		long long session_total_len = 0;
		long long session_total_clicks = 0;
		long long gap = 0;
		long long gap_total_len = 0;
		long long transition = 0;
		long long user_transition = 0;

		if (pos_per_user[u].size() == 0) {
			continue;
		}

		int start = pos_per_user[u].at(0).second;
		int start_ind = 0;
		for (int i = 1; i < (int)pos_per_user[u].size(); i ++) {
			// user transition
			set<string>& owners_cur = per_item_owners[rItemIds[pos_per_user[u].at(i).first]];
			set<string>& owners_pre = per_item_owners[rItemIds[pos_per_user[u].at(i-1).first]];
			bool flag_u_transition = true;
			for (set<string>::iterator it = owners_cur.begin(); it != owners_cur.end(); it ++) {
				if (owners_pre.find(*it) != owners_pre.end()) {
					flag_u_transition = false;
					break;
				}
			}
			if (flag_u_transition) {
				user_transition ++;
			}

			// new session
			if (pos_per_user[u].at(i).second - pos_per_user[u].at(i - 1).second > timeout) {
				session += 1;
				session_total_len += pos_per_user[u].at(i - 1).second - start;
				session_total_clicks += i - start_ind;

				// sample a session and print out
				if ((u + 1) % 50000 == 0 && i - start_ind > 5) {
					printf("\n\n\n=== user %d: session === \n", u);
					for (int ind = start_ind; ind < i; ind ++) {
						string clicked_item = rItemIds[pos_per_user[u].at(ind).first];
						printf("%s    time: %d, ", clicked_item.c_str(), pos_per_user[u].at(ind).second);
						printf("owners: {");
						for (set<string>::iterator owner_it = per_item_owners[clicked_item].begin(); owner_it != per_item_owners[clicked_item].end(); owner_it ++) {
							printf("%s, ", (*owner_it).c_str());
						}
						printf("}\n");
					}
					printf("\n");
				}

				start = pos_per_user[u].at(i).second;
				start_ind = i;
				gap += 1;
				gap_total_len += pos_per_user[u].at(i).second - pos_per_user[u].at(i - 1).second;
			}
		}

		session += 1;
		session_total_len += pos_per_user[u].back().second - start;
		session_total_clicks += pos_per_user[u].size() - start_ind;
		transition = pos_per_user[u].size() - 1;

		// add to total
		total_sessions += session;
		total_sessions_len += session_total_len;
		total_sessions_clicks += session_total_clicks;
		total_gaps += gap;
		total_gaps_len += gap_total_len;
		total_transitions += transition;
		total_user_transitions += user_transition;
	}
	printf("\n  Avg. #sessions/user = %.2f\n", total_sessions * 1.0 / nUsers);
	printf("\n  Avg. len/session = %.2f hrs\n", total_sessions_len / 3600.0 / total_sessions);
	printf("\n  Avg. #clicks/session = %.2f\n", total_sessions_clicks * 1.0 / total_sessions);
	printf("\n  Avg. len of gap = %.2f hrs\n", total_gaps_len / 3600.0 / total_gaps);
	printf("\n  Percentage of transitions between different users = %.2f\n", total_user_transitions * 1.0 / total_transitions);

	int total_ownerships = 0;
	int total_item_with_multi_owner = 0;
	for (int i = 0; i < nItems; i ++) {
		total_ownerships += per_item_owners[rItemIds[i]].size();
		if (per_item_owners[rItemIds[i]].size() > 1) {
			total_item_with_multi_owner ++;
		}
	}
	printf("\n  Avg. #owners/item = %.2f\n", total_ownerships * 1.0 / nItems);
	printf("\n  #item with multiple owners = %d\n", total_item_with_multi_owner);

	vector<int> proj_per_u(nUsers, 0);
	for (int i = 0; i < nItems; i ++) {
		for (set<string>::iterator it = per_item_owners[rItemIds[i]].begin(); it != per_item_owners[rItemIds[i]].end(); it ++) {
			if (userIds.find(*it) != userIds.end()) {
				proj_per_u[userIds[*it]] ++;
			} else {
				printf("owner not found!!!");
				exit(1);
			}
		}
	}
	int total_proj = 0;
	int count_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (proj_per_u[u] > 0) {
			count_user ++;
		}
		total_proj += proj_per_u[u];
	}
	printf("\n  #user with projs = %d\n", count_user);
	printf("\n  Avg. #projs/owner = %.2f\n", total_proj * 1.0 / count_user);
	printf("\n  Avg. #projs/user = %.2f\n", total_proj * 1.0 / nUsers);
}

void corpus::loadOwner(const char* ownerPath)
{
	igzstream in;
	in.open(ownerPath);
	if (! in.good()) {
		printf("Can't read ownerships from %s.\n", ownerPath);
		exit(1);
	}

	string line;
	string item, owner;
	while (getline(in, line)) {
		stringstream ss(line);
		ss >> item >> owner;
		if (per_item_owners.find(item) == per_item_owners.end()) {
			per_item_owners[item] = set<string>();
		}
		per_item_owners[item].insert(owner);
	}
}

void corpus::loadClicks(const char* clickFile, int userMin)
{
	unordered_map<string, int> uCounts;
	unordered_map<string, int> iCounts;

	printf("  Loading clicks from %s, userMin = %d  ", clickFile, userMin);

	string uName; // User name
	string iName; // Item name
	int voteTime; // Time rating was entered

	int nRead = 0; // Progress
	string line;

	igzstream in;
	in.open(clickFile);
	if (! in.good()) {
		printf("Can't read clicks from %s.\n", clickFile);
		exit(1);
	}

	// The first pass collects statsitics
	while (getline(in, line)) {
		stringstream ss(line);
		ss >> uName >> iName >> voteTime;

		nRead ++;
		if (nRead % 100000 == 0) {
			printf(".");
			fflush(stdout);
		}

		if (uCounts.find(uName) == uCounts.end()) {
			uCounts[uName] = 0;
		}
		if (iCounts.find(iName) == iCounts.end()) {
			iCounts[iName] = 0;
		}
		uCounts[uName] ++;
		iCounts[iName] ++;
	}
	in.close();
	printf("\n  First pass: #users = %d, #items = %d, #clicks = %d\n", (int)uCounts.size(), (int)iCounts.size(), nRead);

	// collect all owners
	unordered_set<string> all_owners;
	for (unordered_map<string, int>::iterator it = iCounts.begin(); it != iCounts.end(); it ++) {
		string item = it->first;
		if (per_item_owners.find(item) == per_item_owners.end()) {
			printf("\n  Item %s has no owner.\n", item.c_str());
			exit(1);
		}
		for (set<string>::iterator it = per_item_owners[item].begin(); it != per_item_owners[item].end(); it ++) {
			all_owners.insert(*it);
		}
	}

	// The second pass is for filtering
	nUsers = 0;
	nItems = 0;
	nClicks = 0;

	igzstream in2;
	in2.open(clickFile);
	if (! in2.good()) {
		printf("Can't read clicks from %s.\n", clickFile);
		exit(1);
	}

	nRead = 0;
	while (getline(in2, line)) {
		stringstream ss(line);
		ss >> uName >> iName >> voteTime;

		nRead ++;
		if (nRead % 100000 == 0) {
			printf(".");
			fflush(stdout);
		}

		// validate user---valid := is owner or clicked >userMin times  
		if (uCounts[uName] < userMin && all_owners.find(uName) == all_owners.end()) {
			continue;
		}

		nClicks ++;

		// new item
		if (itemIds.find(iName) == itemIds.end()) {
			rItemIds[nItems] = iName;
			itemIds[iName] = nItems ++;
		}

		// new user
		if (userIds.find(uName) == userIds.end()) {
			rUserIds[nUsers] = uName;
			userIds[uName] = nUsers ++;
			// add a new vec
			pos_per_user.push_back(vector<pair<int,int> >());
		}
		pos_per_user[userIds[uName]].push_back(make_pair(itemIds[iName], voteTime));
	}
	in2.close();

	printf("\n  Filtered inactive users: #users = %d, #items = %d, #clicks = %d\n", nUsers, nItems, nClicks);
	
	// adding users without clicks but being the owners
	for (unordered_set<string>::iterator it = all_owners.begin(); it != all_owners.end(); it ++) {
		if (userIds.find(*it) == userIds.end()) {
			rUserIds[nUsers] = *it;
			userIds[*it] = nUsers ++;
			// add a new vec
			pos_per_user.push_back(vector<pair<int,int> >());
		}
	}
	printf("\n  Added missing owners: #users = %d, #items = %d, #clicks = %d\n", nUsers, nItems, nClicks);

	// rank clicks for each user in terms of timestamps
	printf("\n  Sorting clicks for each users ");
	pos_per_user.shrink_to_fit();
	
	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		pos_per_user[u].shrink_to_fit();
		sort(pos_per_user[u].begin(), pos_per_user[u].end(), timeCompare);
		if (u % 10000 == 0) {
			printf(".");
			fflush(stdout);
		}
	}
	printf("\n");
}

void corpus::loadImgFeatures(const char* imgFeatPath)
{
	imageFeatures = vector<vector<pair<short, float> > >(nItems, vector<pair<short, float> >());

	FILE* f = fopen_(imgFeatPath, "rb");
	printf("\n  Loading image features from %s  ", imgFeatPath);

	float ma = 58.388599; // Largest feature observed

	float* feat = new float [imFeatureDim];
	char* asin = new char [9];
	asin[8] = '\0';
	int a;
	int nRead = 0;
	while (!feof(f)) {
		if ((a = fread(asin, sizeof(*asin), 8, f)) != 8) {
			//printf("Expected to read %d chars, got %d\n", 10, a);
			continue;
		}
		
		string sAsin(asin);

		if ((a = fread(feat, sizeof(*feat), imFeatureDim, f)) != imFeatureDim) {
			printf("Expected to read %d floats, got %d\n", imFeatureDim, a);
			exit(1);
		}
		
		if (itemIds.find(sAsin) == itemIds.end()) {
			continue;
		}

		vector<pair<short, float> > &vec = imageFeatures.at(itemIds[sAsin]);
		for (short f = 0; f < imFeatureDim; f ++) {
			if (feat[f] != 0) {  // compression
				vec.push_back(make_pair(f, feat[f]/ma));
			}
		}
		vec.shrink_to_fit(); // save memory

		// print process
		nRead ++;
		if (not (nRead % 10000)) {
			printf(".");
			fflush(stdout);
		}
	}
	printf("\n");

	delete [] asin;
	delete [] feat;
	fclose(f);
}
