#include "corpus.hpp"
#include "Vista.hpp"
#include "Vistaplus.hpp"


void go_Vista(corpus* corp, int K, int KK, int L, double lambda, int iterations, const char* modelPath) 
{
	Vista md(corp, K, KK, L, lambda);
	md.init();
	md.train(iterations, 0.05);
	md.saveModel((string(modelPath) + "__" + md.toString() + ".txt").c_str());
	md.cleanUp();
}

void go_Vistaplus(corpus* corp, int K, int K2, int L, double lambda, double lambda2, int iterations, const char* modelPath) 
{
	Vistaplus md(corp, K, K2, L, lambda, lambda2);
	md.init();
	md.train(iterations, 0.05);
	md.saveModel((string(modelPath) + "__" + md.toString() + ".txt").c_str());
	md.cleanUp();
}


int main(int argc, char** argv)
{
	srand(0);
	
	if (argc != 12) {
		printf(" Parameters as following: \n");
		printf(" 1. Click/Appreciate triples file path \n");
		printf(" 2. Image feature file path \n");
		printf(" 3. Ownership file path \n");
		printf(" 4. User min (minimum #actions of each user, set to \"5\" in RecSys paper) \n");
		printf(" 5. Dimensionality of vectors measuring \"user-item\" affinity (set to 10 in RecSys paper\n");
		printf(" 6. Dimensionality of vectors measuring \"user-user\"/\"item-item\" affinity (set to 10 in RecSys paper) \n");
		printf(" 7. Order of Markov Chain (1, 2, or 3 in RecSys paper)\n");
		printf(" 8. lambda (l2-norm regularizer)\n");
		printf(" 9. lambda2 (l2-norm regularizer of the embedding matrix)\n");
		printf("10. #Iterations \n");
		printf("11. model path (prefix of the file name to save the model)\n\n");

		exit(1);
	}

	char* reviewPath = argv[1];
	char* imgFeatPath = argv[2];
	char* ownerPath = argv[3];
	int usermin = atoi(argv[4]);
	int K  = atoi(argv[5]);
	int K2 = atoi(argv[6]);
	int L = atoi(argv[7]);
	double lambda = atof(argv[8]);
	double lambda2 = atof(argv[9]);
	int iter = atoi(argv[10]);
	char* modelPath = argv[11];

	printf("{\n");
	printf("  \"corpus\": \"%s\",\n", reviewPath);

	corpus corp;
	corp.loadData(reviewPath, imgFeatPath, ownerPath, usermin);

	// go_Vista(&corp, K, K2, L, lambda, iter, modelPath);
	go_Vistaplus(&corp, K, K2, L, lambda, lambda2, iter, modelPath);
	
	printf("}\n");

	return 0;
}
