#include "model.hpp"

void model::AUC(double* AUC_val, double* AUC_test, double* var)
{
	vector<double> AUC_u_val(nUsers, -1);
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;
		int item_val_prev 	= val_per_user[u].second;

		if (item_test == -1) {  // invalid user
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);
		double x_u_val  = prediction(u, item_val_prev, item_val);

		int count_val = 0;
		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_val_prev, j);
			if (x_u_val > x_uj) {
				count_val ++;
			}

			x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_val[u] = 1.0 * count_val / max;
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_val = 0;
	*AUC_test = 0;
	int num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_val += AUC_u_val[u];
			*AUC_test += AUC_u_test[u];
			num_user ++;
		}
	}
	*AUC_val /= num_user;
	*AUC_test /= num_user;

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / num_user;
}

void model::sampleAUC(double* AUC_val, double* AUC_test, double* var)
{
	vector<double> AUC_u_val(nUsers, -1);
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;
		int item_val_prev 	= val_per_user[u].second;

		if (item_test == -1) {  // invalid user
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);
		double x_u_val  = prediction(u, item_val_prev, item_val);

		int count_val = 0;
		int count_test = 0;
		int max = 0;
		for (int ind = 0; ind < 2000; ind ++) {
			int j = rand() % nItems;

			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_val_prev, j);
			if (x_u_val > x_uj) {
				count_val ++;
			}

			x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_val[u] = 1.0 * count_val / max;
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_val = 0;
	*AUC_test = 0;
	int num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_val += AUC_u_val[u];
			*AUC_test += AUC_u_test[u];
			num_user ++;
		}
	}
	*AUC_val /= num_user;
	*AUC_test /= num_user;

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / num_user;
}

void model::AUC_coldItem(double* AUC_test, double* var, int* num_user)
{
	vector<int> num_pos_per_item(nItems, 0);
	for (int u = 0; u < nUsers; u ++) {
		for (unsigned i = 0; i < corp->pos_per_user[u].size(); i ++) {
			int item = corp->pos_per_user[u][i].first;
			num_pos_per_item[item] += 1;
		}
	}

	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (num_pos_per_item[item_test] >= 5) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void model::AUC_coldUser(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (corp->pos_per_user[u].size() >= 5) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void model::copyBestModel()
{
	for (int w = 0; w < NW; w ++) {
		bestW[w] = W[w];
	}
}

void model::saveModel(const char* path)
{
	FILE* f = fopen_(path, "w");
	fprintf(f, "{\n");
	fprintf(f, "  \"NW\": %d,\n", NW);

	fprintf(f, "  \"W\": [");
	for (int w = 0; w < NW; w ++) {
		fprintf(f, "%f", bestW[w]);
		if (w < NW - 1) fprintf(f, ", ");
	}
	fprintf(f, "]\n");
	fprintf(f, "}\n");
	fclose(f);

	printf("\nModel saved to %s.\n", path);
}

/// model must be first initialized before calling this function
void model::loadModel(const char* path)
{
	printf("\n  loading parameters from %s.\n", path);
	ifstream in;
	in.open(path);
	if (! in.good()){
		printf("Can't read init solution from %s.\n", path);
		exit(1);
	}
	string line;
	string st;
	char ch;
	while(getline(in, line)) {
		stringstream ss(line);
		ss >> st;
		if (st == "\"NW\":") {
			int nw;
			ss >> nw;
			if (nw != NW) {
				printf("NW not match.");
				exit(1);
			}
			continue;
		}

		if (st == "\"W\":") {
			ss >> ch; // skip '['
			for (int w = 0; w < NW; w ++) {
				if (! (ss >> W[w] >> ch)) {
					printf("Read W[] error.");
					exit(1);
				}
			}
			break;
		}
	}
	in.close();
}

void model::AUC_coldUserItem(double* AUC_test, double* var, int* num_user)
{
	vector<vector<int> > item_owners = vector<vector<int> >(nItems, vector<int>());
	for (int i = 0; i < nItems; i ++) {
		string iName = corp->rItemIds[i];
		set<string>& owners = corp->per_item_owners[iName];

		for (set<string>::iterator it = owners.begin(); it != owners.end(); it ++) {
			item_owners[i].push_back(corp->userIds[*it]);
		}
	}

	vector<int> num_pos_per_item(nItems, 0);
	for (int u = 0; u < nUsers; u ++) {
		for (unsigned i = 0; i < corp->pos_per_user[u].size(); i ++) {
			int item = corp->pos_per_user[u][i].first;
			num_pos_per_item[item] += 1;
		}
	}

	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		double avg_owner_click = 0;
		for (vector<int>::iterator it = item_owners[item_test].begin(); it != item_owners[item_test].end(); it ++) {
			avg_owner_click += corp->pos_per_user[*it].size();
		}
		avg_owner_click /= item_owners[item_test].size();
		
		if (num_pos_per_item[item_test] + avg_owner_click >= 10) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

string model::toString()
{
	return "Empty Model!";
}

void model::AUC_OwnerTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		bool owner_tran = true;
		for (set<string>::iterator i = corp->per_item_owners[corp->rItemIds[item_test]].begin(); i != corp->per_item_owners[corp->rItemIds[item_test]].end(); i ++) {
			for (set<string>::iterator j = corp->per_item_owners[corp->rItemIds[item_test_prev]].begin(); j != corp->per_item_owners[corp->rItemIds[item_test_prev]].end(); j ++) {
				if (*i == *j) {
					owner_tran = false;
					break;
				}
			}
		}

		if (! owner_tran) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void model::AUC_NonOwnerTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		bool owner_tran = true;
		for (set<string>::iterator i = corp->per_item_owners[corp->rItemIds[item_test]].begin(); i != corp->per_item_owners[corp->rItemIds[item_test]].end(); i ++) {
			for (set<string>::iterator j = corp->per_item_owners[corp->rItemIds[item_test_prev]].begin(); j != corp->per_item_owners[corp->rItemIds[item_test_prev]].end(); j ++) {
				if (*i == *j) {
					owner_tran = false;
					break;
				}
			}
		}

		if (owner_tran) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void model::AUC_SessionTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (teststamp_per_user[u].first <= teststamp_per_user[u].second + 3600) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void model::AUC_NonSessionTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (teststamp_per_user[u].first > teststamp_per_user[u].second + 3600) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void model::AUC_coldFull(double* AUC_test, double* var, int* num_user)
{
	vector<vector<int> > item_owners = vector<vector<int> >(nItems, vector<int>());
	for (int i = 0; i < nItems; i ++) {
		string iName = corp->rItemIds[i];
		set<string>& owners = corp->per_item_owners[iName];

		for (set<string>::iterator it = owners.begin(); it != owners.end(); it ++) {
			item_owners[i].push_back(corp->userIds[*it]);
		}
	}

	vector<int> num_pos_per_item(nItems, 0);
	vector<int> num_pos_per_owner(nUsers, 0);
	for (int u = 0; u < nUsers; u ++) {
		for (unsigned i = 0; i < corp->pos_per_user[u].size(); i ++) {
			int item = corp->pos_per_user[u][i].first;
			num_pos_per_item[item] += 1;

			for (vector<int>::iterator it = item_owners[item].begin(); it != item_owners[item].end(); it ++) {
				num_pos_per_owner[*it] += 1;
			}
		}
	}

	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		double avg_owner_click = 0;
		double avg_owner_clicked = 0;
		for (vector<int>::iterator it = item_owners[item_test].begin(); it != item_owners[item_test].end(); it ++) {
			avg_owner_click   += corp->pos_per_user[*it].size();
			avg_owner_clicked += num_pos_per_owner[*it];
		}
		avg_owner_click /= item_owners[item_test].size();
		avg_owner_clicked /= item_owners[item_test].size();
		
		if (num_pos_per_item[item_test] + avg_owner_click + avg_owner_clicked >= 10) {
			continue;
		}

		double x_u_test = prediction(u, item_test_prev, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, item_test_prev, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}