#include "Vista.hpp"

void Vista::init()
{
	// init
	item_owners = vector<vector<int> >(nItems, vector<int>());

	for (int i = 0; i < nItems; i ++) {
		string iName = corp->rItemIds[i];
		set<string>& owners = corp->per_item_owners[iName];

		for (set<string>::iterator it = owners.begin(); it != owners.end(); it ++) {
			item_owners[i].push_back(corp->userIds[*it]);
		}
	}

	NW = (K + KK) * (nUsers + nItems) + 2 * nUsers + 2;
	W = new double [NW];
	bestW = new double [NW];

	getParametersFromVector(W, &decay_a, &decay_b, &decay_user_a, &decay_user_b, &gamma_user, &gamma_item, &kappa_user, &kappa_item, INIT);

	for (int w = 0; w < NW; w ++) {
		W[w] = rand() * 1.0 / RAND_MAX;
	}

	*decay_a = 1;
	*decay_b = 0;
	for (int u = 0; u < nUsers; u ++) {
		decay_user_a[u] = 0;
		decay_user_b[u] = 0;
	}
}

void Vista::cleanUp()
{
	getParametersFromVector(W, &decay_a, &decay_b, &decay_user_a, &decay_user_b, &gamma_user, &gamma_item, &kappa_user, &kappa_item, FREE);
	
	delete [] W;
	delete [] bestW;
}

void Vista::getParametersFromVector(	double*   g,
										double**  decay_a,
										double**  decay_b,
										double**  decay_user_a,
										double**  decay_user_b,
										double*** gamma_user, 
										double*** gamma_item,
										double*** kappa_user,
										double*** kappa_item,
										action_t  action)
{
	if (action == FREE) {
		delete [] (*gamma_user);
		delete [] (*gamma_item);
		delete [] (*kappa_user);
		delete [] (*kappa_item);
		return;
	}

	if (action == INIT)	{
		*gamma_user = new double* [nUsers];
		*gamma_item = new double* [nItems];
		*kappa_user = new double* [nUsers];
		*kappa_item = new double* [nItems];
	}

	int ind = 0;

	*decay_a = g + ind;
	ind ++;
	*decay_b = g + ind;
	ind ++;

	*decay_user_a = g + ind;
	ind += nUsers;

	*decay_user_b = g + ind;
	ind += nUsers;

	for (int u = 0; u < nUsers; u ++) {
		(*gamma_user)[u] = g + ind;
		ind += K;
	}
	for (int i = 0; i < nItems; i ++) {
		(*gamma_item)[i] = g + ind;
		ind += K;
	}
	for (int u = 0; u < nUsers; u ++) {
		(*kappa_user)[u] = g + ind;
		ind += KK;
	}
	for (int i = 0; i < nItems; i ++) {
		(*kappa_item)[i] = g + ind;
		ind += KK;
	}

	if (ind != NW) {
		printf("Got bad index (Vista.cpp, line %d)", __LINE__);
		exit(1);
	}
}

double Vista::prediction(int user, vector<int>& prev_items, int item)
{
	double res = inner(gamma_user[user], gamma_item[item], K);
	double inner_a = 0;
	for (vector<int>::iterator it = item_owners[item].begin(); it != item_owners[item].end(); it ++) {
		inner_a += inner(kappa_user[user], kappa_user[*it], KK);
	}
	inner_a /= item_owners[item].size();
	res += inner_a;

	// iterate through prev items
	for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
		int cur_prev_item = prev_items[ind];
		double decay_wt = exp(-(*decay_a + decay_user_a[user]) * ind - (*decay_b + decay_user_b[user]));

		double inner_b = 0;
		for (vector<int>::iterator i = item_owners[item].begin(); i != item_owners[item].end(); i ++) {
			for (vector<int>::iterator j = item_owners[cur_prev_item].begin(); j != item_owners[cur_prev_item].end(); j ++) {
				inner_b += inner(kappa_user[*i], kappa_user[*j], KK);
			}
		}
		inner_b /= item_owners[item].size() * item_owners[cur_prev_item].size();

		res += decay_wt * (inner_b + inner(kappa_item[item], kappa_item[cur_prev_item], KK));
	}
	return res;
}

int Vista::sampleUser()
{
	while (true) {
		int user_id = rand() % nUsers;
		if (corp->pos_per_user[user_id].size() < 2) {
			continue;
		}
		return user_id;
	}
}

void Vista::updateFactors(int user_id, vector<int>& prev_items, int pos_item_id, int neg_item_id, double learn_rate)
{
	double* gpos_minus_gneg = new double [K];
	double* kpos_minus_kneg = new double [KK];
	double* user_plus_ui = new double [KK];
	double* upos_minus_uneg = new double [KK];
	double* kappa_sum = new double [KK];

	int owner_pos = item_owners[pos_item_id][rand() % item_owners[pos_item_id].size()];
	int owner_neg = item_owners[neg_item_id][rand() % item_owners[neg_item_id].size()];
	vector<int> owner_curs; 
	for (int i = 0; i < (int)prev_items.size(); i ++) {
		owner_curs.push_back(item_owners[prev_items[i]][rand() % item_owners[prev_items[i]].size()]);
	}

	vector<double> decay_wts(prev_items.size(), 1);
	for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
		decay_wts[ind] = exp(- (*decay_a + decay_user_a[user_id]) * ind - (*decay_b + decay_user_b[user_id]));
	}

	for (int k = 0; k < K; k ++) {
		gpos_minus_gneg[k] = gamma_item[pos_item_id][k] - gamma_item[neg_item_id][k];
	}
	for (int k = 0; k < KK; k ++) {
		upos_minus_uneg[k] = kappa_user[owner_pos][k] - kappa_user[owner_neg][k];
		kpos_minus_kneg[k] = kappa_item[pos_item_id][k] - kappa_item[neg_item_id][k];

		user_plus_ui[k] = kappa_user[user_id][k];
		kappa_sum[k] = 0;
		for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
			user_plus_ui[k] += decay_wts[ind] * kappa_user[owner_curs[ind]][k];
			kappa_sum[k] += decay_wts[ind] * kappa_item[prev_items[ind]][k];
		}
	}

	double x_uij = inner(gamma_user[user_id], gpos_minus_gneg, K) 
				 + inner(user_plus_ui, upos_minus_uneg, KK)
				 + inner(kappa_sum, kpos_minus_kneg, KK);

	double deri = 1.0 / (1.0 + exp(x_uij));

	// adjust latent factors
	for (int f = 0; f < K; f ++) {
		double deri_A = deri * gamma_user[user_id][f];

		gamma_user[user_id][f]     += learn_rate * ( deri * gpos_minus_gneg[f] - lambda * gamma_user[user_id][f]);
		gamma_item[pos_item_id][f] += learn_rate * ( deri_A - lambda * gamma_item[pos_item_id][f]);
		gamma_item[neg_item_id][f] += learn_rate * (-deri_A - lambda / 10.0 * gamma_item[neg_item_id][f]);
	}

	for (int f = 0; f < KK; f ++) {
		double deri_B = deri * user_plus_ui[f];
		double deri_D = deri * upos_minus_uneg[f];
		double deri_E = deri * kappa_sum[f];

		kappa_user[user_id][f]   += learn_rate * ( deri_D - lambda * kappa_user[user_id][f]);
		kappa_user[owner_pos][f] += learn_rate * ( deri_B - lambda * kappa_user[owner_pos][f]);
		kappa_user[owner_neg][f] += learn_rate * (-deri_B - lambda / 10.0 * kappa_user[owner_neg][f]);

		kappa_item[pos_item_id][f] += learn_rate * ( deri_E - lambda * kappa_item[pos_item_id][f]);
		kappa_item[neg_item_id][f] += learn_rate * (-deri_E - lambda / 10.0 * kappa_item[neg_item_id][f]);

		for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
			int cur_prev_item = prev_items[ind];
			int cur_prev_owner = owner_curs[ind];

			kappa_user[cur_prev_owner][f] += learn_rate * (deri_D * decay_wts[ind] - lambda * kappa_user[cur_prev_owner][f]);
			kappa_item[cur_prev_item][f]  += learn_rate * (deri * decay_wts[ind] * kpos_minus_kneg[f] - lambda * kappa_item[cur_prev_item][f]);
		}
	}

	for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
		double deri_decay = deri * decay_wts[ind] * 
			 	( inner(kappa_item[prev_items[ind]], kpos_minus_kneg, KK) 
			 	+ inner(kappa_user[owner_curs[ind]], upos_minus_uneg, KK) );

		// You may want to tune the discounted factor for learning rate (i.e. 1000) a bit
		*decay_a += learn_rate / 1000 * (deri_decay * (-ind));  
		*decay_b += learn_rate / 1000 * (deri_decay * (-1));
		decay_user_a[user_id] += learn_rate / 1000 * (deri_decay * (-ind));
		decay_user_b[user_id] += learn_rate / 1000 * (deri_decay * (-1));
	}
	*decay_a -= learn_rate * lambda * (*decay_a);
	*decay_b -= learn_rate * lambda * (*decay_b);
	decay_user_a[user_id] -= learn_rate * lambda * decay_user_a[user_id];
	decay_user_b[user_id] -= learn_rate * lambda * decay_user_b[user_id];

	delete [] gpos_minus_gneg;
	delete [] kpos_minus_kneg;
	delete [] user_plus_ui;
	delete [] upos_minus_uneg;
	delete [] kappa_sum;
}

void Vista::oneiteration(double learn_rate)
{
	// working memory
	vector<pair<int, vector<int> > >* user_matrix = new vector<pair<int, vector<int> > > [nUsers];
	for (int u = 0; u < nUsers; u ++) {
		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];
		for (int i = 1; i < (int)user_pos.size(); i ++) {
			vector<int> prev_items;
			for (int j = i - 1; j >= max(0, i - L); j --) {
				prev_items.push_back(user_pos[j].first);
			} 
			user_matrix[u].push_back(make_pair(user_pos[i].first, prev_items));
		}
	}

	// now it begins!
	#pragma omp parallel for schedule(dynamic)
	for (int i = 0; i < num_pos_events; i ++) {
		int user_id, pos_item_id, neg_item_id;
		vector<int> prev_items;

		// sample user
		user_id = sampleUser();
		vector<pair<int, vector<int> > >& user_items = user_matrix[user_id];

		// sample positive item
		int rand_num = rand() % user_items.size();
		pos_item_id = user_items.at(rand_num).first;
		prev_items = user_items.at(rand_num).second;

		// sample negative item
		do {
			neg_item_id = rand() % nItems;
		} while (neg_item_id == pos_item_id);

		// now got tuple (user_id, pos_item, neg_item)
		updateFactors(user_id, prev_items, pos_item_id, neg_item_id, learn_rate);
	}

	delete [] user_matrix;
}

void Vista::train(int iterations, double learn_rate)
{
	printf("%s", ("\n<<< " + toString() + " >>>\n\n").c_str());

	double bestValidAUC = -1;
	int best_iter = 0;

	// SGD begins
	for (int iter = 1; iter <= iterations; iter ++) {
		
		// perform one iter of SGD
		double l_dlStart = clock_();
		oneiteration(learn_rate);
		printf("Iter: %d, took %f\n", iter, clock_() - l_dlStart);
		fflush(stdout);

		if (iter % 10 == 0) {
			double valid, test, var;
			sampleAUC(&valid, &test, &var);
			printf("[Valid AUC = %f], Test AUC = %f, Test var = %f\n", valid, test, var);
			fflush(stdout);
			
			if (bestValidAUC < valid) {
				bestValidAUC = valid;
				best_iter = iter;
				copyBestModel();
				saveModel((string("model.") + toString() + string(".snapshot.txt")).c_str());
			} else if (iter > best_iter + 40) {
				printf("Overfitted. Exiting... \n");
				break;
			}
		}
	}

	// copy back best parameters
	for (int w = 0; w < NW; w ++) {
		W[w] = bestW[w];
	}

	double valid, test, var;
	int num_item;

	AUC(&valid, &test, &var);
	printf("\n\n <<< %s >>> Test AUC = %f, Test var = %f\n", toString().c_str(), test, var);

	AUC_coldUser(&test, &var, &num_item);
	printf("\n\n <<< %s >>> User Cold Start: #User = %d, Test AUC = %f, Test Var = %f\n", toString().c_str(), num_item, test, var);

	AUC_coldItem(&test, &var, &num_item);
	printf("\n\n <<< %s >>> Item Cold Start: #Item = %d, Test AUC = %f, Test var = %f\n", toString().c_str(), num_item, test, var);

	AUC_coldUserItem(&test, &var, &num_item);
	printf("\n\n <<< %s >>> UserItem Cold Start: #User = %d, Test AUC = %f, Test Var = %f\n", toString().c_str(), num_item, test, var);
	
	AUC_OwnerTransition(&test, &var, &num_item);
	printf("\n\n <<< %s >>> Owner-Transition: #case = %d, Test AUC = %f, Test var = %f\n", toString().c_str(), num_item, test, var);
	
	AUC_NonOwnerTransition(&test, &var, &num_item);
	printf("\n\n <<< %s >>> Non Owner-Transition: #case = %d, Test AUC = %f, Test var = %f\n", toString().c_str(), num_item, test, var);
	
	AUC_SessionTransition(&test, &var, &num_item);
	printf("\n\n <<< %s >>> Session Transition: #case = %d, Test AUC = %f, Test var = %f\n", toString().c_str(), num_item, test, var);
	
	AUC_NonSessionTransition(&test, &var, &num_item);
	printf("\n\n <<< %s >>> Non Session-Transition: #case = %d, Test AUC = %f, Test var = %f\n", toString().c_str(), num_item, test, var);
}

string Vista::toString()
{
	char str[100];
	sprintf(str, "Vista__K_%d_L_%d_lambda_%.2f", K, L, lambda);
	return str;
}

void Vista::AUC(double* AUC_val, double* AUC_test, double* var)
{
	vector<double> AUC_u_val(nUsers, -1);
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test = test_per_user[u].first;
		int item_val  = val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}

		vector<int> val_prev_items;
		for (int i = 1; i <= L; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			val_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}
		
		double x_u_test = prediction(u, test_prev_items, item_test);
		double x_u_val  = prediction(u, val_prev_items, item_val);

		int count_val = 0;
		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, val_prev_items, j);
			if (x_u_val > x_uj) {
				count_val ++;
			}

			x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_val[u] = 1.0 * count_val / max;
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_val = 0;
	*AUC_test = 0;
	int num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_val += AUC_u_val[u];
			*AUC_test += AUC_u_test[u];
			num_user ++;
		}
	}
	*AUC_val /= num_user;
	*AUC_test /= num_user;

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / num_user;
}

void Vista::sampleAUC(double* AUC_val, double* AUC_test, double* var)
{
	vector<double> AUC_u_val(nUsers, -1);
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test = test_per_user[u].first;
		int item_val  = val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}

		vector<int> val_prev_items;
		for (int i = 1; i <= L; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			val_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}
		
		double x_u_test = prediction(u, test_prev_items, item_test);
		double x_u_val  = prediction(u, val_prev_items, item_val);

		int count_val = 0;
		int count_test = 0;
		int max = 0;
		for (int ind = 0; ind < 2000; ind ++) {
			int j = rand() % nItems;

			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, val_prev_items, j);
			if (x_u_val > x_uj) {
				count_val ++;
			}

			x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_val[u] = 1.0 * count_val / max;
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_val = 0;
	*AUC_test = 0;
	int num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_val += AUC_u_val[u];
			*AUC_test += AUC_u_test[u];
			num_user ++;
		}
	}
	*AUC_val /= num_user;
	*AUC_test /= num_user;

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / num_user;
}

void Vista::AUC_coldItem(double* AUC_test, double* var, int* num_user)
{
	vector<int> num_pos_per_item(nItems, 0);
	for (int u = 0; u < nUsers; u ++) {
		for (unsigned i = 0; i < corp->pos_per_user[u].size(); i ++) {
			int item = corp->pos_per_user[u][i].first;
			num_pos_per_item[item] += 1;
		}
	}

	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test = test_per_user[u].first;
		int item_val  = val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (num_pos_per_item[item_test] >= 5) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void Vista::AUC_coldUser(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test = test_per_user[u].first;
		int item_val  = val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (corp->pos_per_user[u].size() >= 5) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}


void Vista::AUC_coldUserItem(double* AUC_test, double* var, int* num_user)
{
	vector<int> num_pos_per_item(nItems, 0);
	for (int u = 0; u < nUsers; u ++) {
		for (unsigned i = 0; i < corp->pos_per_user[u].size(); i ++) {
			int item = corp->pos_per_user[u][i].first;
			num_pos_per_item[item] += 1;
		}
	}

	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test = test_per_user[u].first;
		int item_val  = val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		double avg_owner_click = 0;
		for (vector<int>::iterator it = item_owners[item_test].begin(); it != item_owners[item_test].end(); it ++) {
			avg_owner_click += corp->pos_per_user[*it].size();
		}
		avg_owner_click /= item_owners[item_test].size();
		
		if (num_pos_per_item[item_test] + avg_owner_click >= 10) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void Vista::AUC_OwnerTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		bool owner_tran = true;
		for (set<string>::iterator i = corp->per_item_owners[corp->rItemIds[item_test]].begin(); i != corp->per_item_owners[corp->rItemIds[item_test]].end(); i ++) {
			for (set<string>::iterator j = corp->per_item_owners[corp->rItemIds[item_test_prev]].begin(); j != corp->per_item_owners[corp->rItemIds[item_test_prev]].end(); j ++) {
				if (*i == *j) {
					owner_tran = false;
					break;
				}
			}
		}

		if (! owner_tran) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void Vista::AUC_NonOwnerTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_test_prev  = test_per_user[u].second;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		bool owner_tran = true;
		for (set<string>::iterator i = corp->per_item_owners[corp->rItemIds[item_test]].begin(); i != corp->per_item_owners[corp->rItemIds[item_test]].end(); i ++) {
			for (set<string>::iterator j = corp->per_item_owners[corp->rItemIds[item_test_prev]].begin(); j != corp->per_item_owners[corp->rItemIds[item_test_prev]].end(); j ++) {
				if (*i == *j) {
					owner_tran = false;
					break;
				}
			}
		}

		if (owner_tran) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void Vista::AUC_SessionTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (teststamp_per_user[u].first <= teststamp_per_user[u].second + 3600) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void Vista::AUC_NonSessionTransition(double* AUC_test, double* var, int* num_user)
{
	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test 		= test_per_user[u].first;
		int item_val  		= val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		if (teststamp_per_user[u].first > teststamp_per_user[u].second + 3600) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}

void Vista::AUC_coldFull(double* AUC_test, double* var, int* num_user)
{
	vector<int> num_pos_per_item(nItems, 0);
	vector<int> num_pos_per_owner(nUsers, 0);
	for (int u = 0; u < nUsers; u ++) {
		for (unsigned i = 0; i < corp->pos_per_user[u].size(); i ++) {
			int item = corp->pos_per_user[u][i].first;
			num_pos_per_item[item] += 1;

			for (vector<int>::iterator it = item_owners[item].begin(); it != item_owners[item].end(); it ++) {
				num_pos_per_owner[*it] += 1;
			}
		}
	}

	vector<double> AUC_u_test(nUsers, -1);

	#pragma omp parallel for schedule(dynamic)
	for (int u = 0; u < nUsers; u ++) {
		int item_test = test_per_user[u].first;
		int item_val  = val_per_user[u].first;

		if (item_test == -1) {  // invalid user
			continue;
		}

		double avg_owner_click = 0;
		double avg_owner_clicked = 0;
		for (vector<int>::iterator it = item_owners[item_test].begin(); it != item_owners[item_test].end(); it ++) {
			avg_owner_click   += corp->pos_per_user[*it].size();
			avg_owner_clicked += num_pos_per_owner[*it];
		}
		avg_owner_click /= item_owners[item_test].size();
		avg_owner_clicked /= item_owners[item_test].size();
		
		if (num_pos_per_item[item_test] + avg_owner_click + avg_owner_clicked >= 10) {
			continue;
		}

		vector<pair<int,int> >& user_pos = corp->pos_per_user[u];

		vector<int> test_prev_items;
		test_prev_items.push_back(test_per_user[u].second);
		for (int i = 1; i <= L - 1; i ++) {
			if ((int)user_pos.size() - i < 0) {
				break;
			}
			test_prev_items.push_back(user_pos[user_pos.size() - i].first);
		}	

		double x_u_test = prediction(u, test_prev_items, item_test);

		int count_test = 0;
		int max = 0;
		for (int j = 0; j < nItems; j ++) {
			if (clicked_per_user[u].find(j) != clicked_per_user[u].end() 
				|| j == item_test
				|| j == item_val) {
				continue;
			}
			max ++;
			double x_uj = prediction(u, test_prev_items, j);
			if (x_u_test > x_uj) {
				count_test ++;
			}
		}
		AUC_u_test[u] = 1.0 * count_test / max;
	}

	// sum up AUC
	*AUC_test = 0;
	*num_user = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			*AUC_test += AUC_u_test[u];
			(*num_user) ++;
		}
	}
	*AUC_test /= (*num_user);

	// calculate standard deviation
	double variance = 0;
	for (int u = 0; u < nUsers; u ++) {
		if (AUC_u_test[u] != -1) {
			variance += square(AUC_u_test[u] - *AUC_test);
		}
	}
	*var = variance / *num_user;
}