#pragma once

#include "model.hpp"

class Vista : public model
{
public:
	Vista(corpus* corp, int K, int KK, int L, double lambda) 
		: model(corp)
		, K(K)
		, KK(KK)
		, L(L)
		, lambda(lambda) {}

	~Vista(){}
	
	void init();
	void cleanUp();

	double prediction(int user, int prev_item, int item) { return -1;}
	double prediction(int user, vector<int>& prev_items, int item);

	void getParametersFromVector(	double*   g,
									double**  decay_a,
									double**  decay_b,
									double**  decay_user_a,
									double**  decay_user_b,
									double*** gamma_user, 
									double*** gamma_item,
									double*** kappa_user,
									double*** kappa_item,
									action_t  action);

	int sampleUser();
	void train(int iterations, double learn_rate);
	void oneiteration(double learn_rate);
	virtual void updateFactors(int user_id, vector<int>& prev_items, int pos_item_id, int neg_item_id, double learn_rate);
	string toString();

	virtual void sampleAUC(double* AUC_val, double* AUC_test, double* var);
	virtual void AUC(double* AUC_val, double* AUC_test, double* var);
	virtual void AUC_coldItem(double* AUC_test, double* var, int* num_user);
	virtual void AUC_coldUser(double* AUC_test, double* var, int* num_user);
	virtual void AUC_coldUserItem(double* AUC_test, double* var, int* num_user);
	virtual void AUC_coldFull(double* AUC_test, double* var, int* num_user);
	
	virtual void AUC_OwnerTransition(double* AUC_test, double* var, int* num_user);
	virtual void AUC_NonOwnerTransition(double* AUC_test, double* var, int* num_user);
	virtual void AUC_SessionTransition(double* AUC_test, double* var, int* num_user);
	virtual void AUC_NonSessionTransition(double* AUC_test, double* var, int* num_user);

	/* auxiliary variables */
	vector<vector<int> > item_owners;

	/* variables */
	double** gamma_user;
	double** gamma_item;
	double** kappa_user;
	double** kappa_item;
	double* decay_user_a;
	double* decay_user_b;
	double* decay_a;
	double* decay_b;

	/* hyper-parameters */
	int K;
	int KK;
	int L;
	double lambda;
};