#include "Vistaplus.hpp"

void Vistaplus::init()
{
	// init
	item_owners = vector<vector<int> >(nItems, vector<int>());

	for (int i = 0; i < nItems; i ++) {
		string iName = corp->rItemIds[i];
		set<string>& owners = corp->per_item_owners[iName];

		for (set<string>::iterator it = owners.begin(); it != owners.end(); it ++) {
			item_owners[i].push_back(corp->userIds[*it]);
		}
	}

	NW = K * nUsers + K * nItems + K * corp->imFeatureDim  // gamma_u, gamma_i, E
		+ KK * nUsers + KK * nItems  // kappa_u, kappa_i
		+ KK * corp->imFeatureDim    // E
		+ 2 * nUsers + 2;            // a, b
	
	W = new double [NW];
	bestW = new double [NW];

	getParametersFromVector(W, &decay_a, &decay_b, &decay_user_a, &decay_user_b, &gamma_user, &gamma_item, &kappa_user, &kappa_item, &E, &Ep, INIT);

	for (int w = 0; w < NW; w ++) {
		W[w] = rand() * 1.0 / RAND_MAX;
	}

	*decay_a = 1;
	*decay_b = 0;
	for (int u = 0; u < nUsers; u ++) {
		decay_user_a[u] = 0;
		decay_user_b[u] = 0;
	}
	
	theta_item = new double* [nItems];
	for (int i = 0; i < nItems; i ++) {
		theta_item[i] = new double [K];
	}

	kappaV_item = new double* [nItems];
	for (int i = 0; i < nItems; i ++) {
		kappaV_item[i] = new double [KK];
	}
}

void Vistaplus::cleanUp()
{
	getParametersFromVector(W, &decay_a, &decay_b, &decay_user_a, &decay_user_b, &gamma_user, &gamma_item, &kappa_user, &kappa_item, &E, &Ep, FREE);
	
	delete [] W;
	delete [] bestW;

	for (int i = 0; i < nItems; i ++) {
		delete [] theta_item[i];
	}
	delete [] theta_item;

	for (int i = 0; i < nItems; i ++) {
		delete [] kappaV_item[i];
	}
	delete [] kappaV_item;
}

void Vistaplus::getParametersFromVector(double*   g,
										double**  decay_a,
										double**  decay_b,
										double**  decay_user_a,
										double**  decay_user_b,
										double*** gamma_user, 
										double*** gamma_item,
										double*** kappa_user,
										double*** kappa_item,
										double*** E,
										double*** Ep,
										action_t  action)
{
	if (action == FREE) {
		delete [] (*gamma_user);
		delete [] (*gamma_item);
		delete [] (*kappa_user);
		delete [] (*kappa_item);
		delete [] (*E);
		delete [] (*Ep);
		return;
	}

	if (action == INIT)	{
		*gamma_user = new double* [nUsers];
		*gamma_item = new double* [nItems];
		*kappa_user = new double* [nUsers];
		*kappa_item = new double* [nItems];
		*E  = new double* [K];
		*Ep = new double* [KK];
	}

	int ind = 0;

	*decay_a = g + ind;
	ind ++;
	*decay_b = g + ind;
	ind ++;

	*decay_user_a = g + ind;
	ind += nUsers;

	*decay_user_b = g + ind;
	ind += nUsers;

	for (int u = 0; u < nUsers; u ++) {
		(*gamma_user)[u] = g + ind;
		ind += K;
	}
	for (int i = 0; i < nItems; i ++) {
		(*gamma_item)[i] = g + ind;
		ind += K;
	}
	
	for (int u = 0; u < nUsers; u ++) {
		(*kappa_user)[u] = g + ind;
		ind += KK;
	}
	for (int i = 0; i < nItems; i ++) {
		(*kappa_item)[i] = g + ind;
		ind += KK;
	}
	for (int k = 0; k < K; k ++) {
		(*E)[k] = g + ind;
		ind += corp->imFeatureDim;
	}
	for (int k = 0; k < KK; k ++) {
		(*Ep)[k] = g + ind;
		ind += corp->imFeatureDim;
	}

	if (ind != NW) {
		printf("Got bad index (Vistaplus.cpp, line %d)", __LINE__);
		exit(1);
	}
}

void Vistaplus::projectX(int x)
{
	vector<pair<short, float> >& feat = corp->imageFeatures.at(x);
	for (int k = 0; k < K; ++ k) {
		theta_item[x][k] = 0;
		for (unsigned i = 0; i < feat.size(); i ++) {
			theta_item[x][k] += E[k][feat[i].first] * feat[i].second;
		}
	}
	for (int k = 0; k < KK; ++ k) {
		kappaV_item[x][k] = 0;
		for (unsigned i = 0; i < feat.size(); i ++) {
			kappaV_item[x][k] += Ep[k][feat[i].first] * feat[i].second;
		}
	}
}

void Vistaplus::project()
{
	#pragma omp parallel for schedule(dynamic)
	for (int x = 0; x < nItems; x ++) {
		projectX(x);
	}
}

double Vistaplus::prediction(int user, vector<int>& prev_items, int item)
{
	double res = inner(gamma_user[user], gamma_item[item], K) + inner(gamma_user[user], theta_item[item], K);
	double inner_a = 0;
	for (vector<int>::iterator it = item_owners[item].begin(); it != item_owners[item].end(); it ++) {
		inner_a += inner(kappa_user[user], kappa_user[*it], KK);
	}
	inner_a /= item_owners[item].size();
	res += inner_a;

	// iterate through prev items
	double* kappa_item_aug = new double [KK];
	for (int k = 0; k < KK; k ++) {
		kappa_item_aug[k] = kappa_item[item][k] + kappaV_item[item][k];
	}

	double* kappa_item_cur_aug = new double [KK];
	for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
		int cur_prev_item = prev_items[ind];
		double decay_wt = exp(-(*decay_a + decay_user_a[user]) * ind - (*decay_b + decay_user_b[user]));

		double inner_b = 0;
		for (vector<int>::iterator i = item_owners[item].begin(); i != item_owners[item].end(); i ++) {
			for (vector<int>::iterator j = item_owners[cur_prev_item].begin(); j != item_owners[cur_prev_item].end(); j ++) {
				inner_b += inner(kappa_user[*i], kappa_user[*j], KK);
			}
		}
		inner_b /= item_owners[item].size() * item_owners[cur_prev_item].size();

		for (int k = 0; k < KK; k ++) {
			kappa_item_cur_aug[k] = kappa_item[cur_prev_item][k] + kappaV_item[cur_prev_item][k];
		}
		res += decay_wt * (inner_b + inner(kappa_item_cur_aug, kappa_item_aug, KK));
	}

	delete [] kappa_item_cur_aug;
	delete [] kappa_item_aug;
	return res;
}

/** NOTE: Here multiple threads are sharing the same theta_item memory, which can be a potential problem. */
void Vistaplus::updateFactors(int user_id, vector<int>& prev_items, int pos_item_id, int neg_item_id, double learn_rate)
{
	double* gpos_minus_gneg = new double [K];	
	double* kpos_minus_kneg = new double [KK];
	double* user_plus_ui = new double [KK];
	double* upos_minus_uneg = new double [KK];
	double* kappa_sum = new double [KK];

	int owner_pos = item_owners[pos_item_id][rand() % item_owners[pos_item_id].size()];
	int owner_neg = item_owners[neg_item_id][rand() % item_owners[neg_item_id].size()];
	vector<int> owner_curs; 
	for (int i = 0; i < (int)prev_items.size(); i ++) {
		projectX(prev_items[i]);
		owner_curs.push_back(item_owners[prev_items[i]][rand() % item_owners[prev_items[i]].size()]);
	}

	vector<double> decay_wts(prev_items.size(), 1);
	for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
		decay_wts[ind] = exp(- (*decay_a + decay_user_a[user_id]) * ind - (*decay_b + decay_user_b[user_id]));
	}

	projectX(pos_item_id);
	projectX(neg_item_id);

	for (int k = 0; k < K; k ++) {
		gpos_minus_gneg[k] = gamma_item[pos_item_id][k] + theta_item[pos_item_id][k] - gamma_item[neg_item_id][k] - theta_item[neg_item_id][k];
	}

	for (int k = 0; k < KK; k ++) {
		kpos_minus_kneg[k] = kappa_item[pos_item_id][k] + kappaV_item[pos_item_id][k] - kappa_item[neg_item_id][k] - kappaV_item[neg_item_id][k];
		upos_minus_uneg[k] = kappa_user[owner_pos][k] - kappa_user[owner_neg][k];

		kappa_sum[k] = 0;
		user_plus_ui[k] = kappa_user[user_id][k];
		for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
			user_plus_ui[k] += decay_wts[ind] * kappa_user[owner_curs[ind]][k];
			kappa_sum[k]  += decay_wts[ind] * (kappa_item[prev_items[ind]][k] + kappaV_item[prev_items[ind]][k]);
		}
	}

	double x_uij = inner(gamma_user[user_id], gpos_minus_gneg, K)
				 + inner(user_plus_ui, upos_minus_uneg, KK)
				 + inner(kappa_sum, kpos_minus_kneg, KK);
	
	double deri = 1.0 / (1.0 + exp(x_uij));

	bool updateE = rand() % 20 == 0; // update the embedding

	vector<pair<int, float> > diff;
	if (updateE) {
		vector<pair<short, float> >& feat_i = corp->imageFeatures[neg_item_id];
		vector<pair<short, float> >& feat_j = corp->imageFeatures[pos_item_id];

		unsigned p_i = 0, p_j = 0;
		while (p_i < feat_i.size() && p_j < feat_j.size()) {
			int ind_i = feat_i[p_i].first;
			int ind_j = feat_j[p_j].first;
			if (ind_i < ind_j) {
				diff.push_back(make_pair(ind_i, - feat_i[p_i].second));
				p_i ++;
			} else if (ind_i > ind_j) {
				diff.push_back(make_pair(ind_j, feat_j[p_j].second));
				p_j ++;
			} else {
				diff.push_back(make_pair(ind_j, feat_j[p_j].second - feat_i[p_i].second));
				p_i ++; p_j ++;
			}
		}
		while (p_i < feat_i.size()) {
			diff.push_back(make_pair(feat_i[p_i].first, - feat_i[p_i].second));
			p_i ++;
		}
		while (p_j < feat_j.size()) {
			diff.push_back(make_pair(feat_j[p_j].first, feat_j[p_j].second));
			p_j ++;
		}

		// update E
		for (int f = 0; f < K; f ++) {
			double deri_E = deri * gamma_user[user_id][f];
			for (unsigned i = 0; i < diff.size(); i ++) {
				E[f][diff[i].first] += learn_rate * (deri_E * diff[i].second - lambda2 * E[f][diff[i].first]);
			}
		}
	}

	// adjust latent factors
	for (int f = 0; f < K; f ++) {
		double deri_A = deri * gamma_user[user_id][f];

		gamma_user[user_id][f]     += learn_rate * ( deri * gpos_minus_gneg[f] - lambda * gamma_user[user_id][f]);
		gamma_item[pos_item_id][f] += learn_rate * ( deri_A - lambda * gamma_item[pos_item_id][f]);
		gamma_item[neg_item_id][f] += learn_rate * (-deri_A - lambda / 10.0 * gamma_item[neg_item_id][f]);
	}

	for (int f = 0; f < KK; f ++) {
		double deri_B = deri * user_plus_ui[f];
		double deri_C = deri * kappa_sum[f];
		double deri_D = deri * upos_minus_uneg[f];

		kappa_user[user_id][f]   += learn_rate * ( deri_D - lambda * kappa_user[user_id][f]);
		kappa_user[owner_pos][f] += learn_rate * ( deri_B - lambda * kappa_user[owner_pos][f]);
		kappa_user[owner_neg][f] += learn_rate * (-deri_B - lambda / 10.0 * kappa_user[owner_neg][f]);

		kappa_item[pos_item_id][f] += learn_rate * ( deri_C - lambda * kappa_item[pos_item_id][f]);
		kappa_item[neg_item_id][f] += learn_rate * (-deri_C - lambda / 10.0 * kappa_item[neg_item_id][f]);

		if (updateE) {
			double deri_E = deri * kappa_sum[f];
			for (unsigned i = 0; i < diff.size(); i ++) {
				Ep[f][diff[i].first] += learn_rate * (deri_E * diff[i].second - lambda2 * Ep[f][diff[i].first]);
			}
		}

		for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
			int cur_prev_item = prev_items[ind];
			int cur_prev_owner = owner_curs[ind];

			kappa_user[cur_prev_owner][f] += learn_rate * (deri_D * decay_wts[ind] - lambda * kappa_user[cur_prev_owner][f]);
			kappa_item[cur_prev_item][f]  += learn_rate * (deri * decay_wts[ind] * kpos_minus_kneg[f] - lambda * kappa_item[cur_prev_item][f]);

			if (updateE) {
				vector<pair<short, float> >& feat = corp->imageFeatures.at(cur_prev_item);
				double deri_G = deri * kpos_minus_kneg[f] * decay_wts[ind];
				for (unsigned i = 0; i < feat.size(); i ++) {
					Ep[f][feat[i].first] += learn_rate * (deri_G * feat[i].second - lambda2 * Ep[f][feat[i].first]);
				}
			}
		}
	}

	for (int ind = 0; ind < (int)prev_items.size(); ind ++) {
		double deri_decay = deri * decay_wts[ind] * 
			 	( inner( kappa_item[prev_items[ind]], kpos_minus_kneg, KK)
			 	+ inner(kappaV_item[prev_items[ind]], kpos_minus_kneg, KK) 
			 	+ inner(kappa_user[owner_curs[ind]], upos_minus_uneg, KK) );

		*decay_a += learn_rate / 1000 * (deri_decay * (-ind));
		*decay_b += learn_rate / 1000 * (deri_decay * (-1));

		decay_user_a[user_id] += learn_rate / 1000 * (deri_decay * (-ind));
		decay_user_b[user_id] += learn_rate / 1000 * (deri_decay * (-1));
	}
	*decay_a -= learn_rate * lambda * (*decay_a);
	*decay_b -= learn_rate * lambda * (*decay_b);
	decay_user_a[user_id] -= learn_rate * lambda * decay_user_a[user_id];
	decay_user_b[user_id] -= learn_rate * lambda * decay_user_b[user_id];

	delete [] gpos_minus_gneg;
	delete [] kpos_minus_kneg;
	delete [] user_plus_ui;
	delete [] upos_minus_uneg;
	delete [] kappa_sum;
}

string Vistaplus::toString()
{
	char str[100];
	sprintf(str, "Vista_plus__K_%d_KK_%d_L_%d_lambda_%.2f", K, KK, L, lambda);
	return str;
}

void Vistaplus::AUC(double* AUC_val, double* AUC_test, double* var)
{
	project();
	Vista::AUC(AUC_val, AUC_test, var);
}

void Vistaplus::sampleAUC(double* AUC_val, double* AUC_test, double* var)
{
	project();
	Vista::sampleAUC(AUC_val, AUC_test, var);
}

void Vistaplus::AUC_coldItem(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_coldItem(AUC_test, var, num_user);
}

void Vistaplus::AUC_coldUser(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_coldUser(AUC_test, var, num_user);
}


void Vistaplus::AUC_coldUserItem(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_coldUserItem(AUC_test, var, num_user);
}


void Vistaplus::AUC_OwnerTransition(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_OwnerTransition(AUC_test, var, num_user);
}

void Vistaplus::AUC_NonOwnerTransition(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_NonOwnerTransition(AUC_test, var, num_user);
}

void Vistaplus::AUC_SessionTransition(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_SessionTransition(AUC_test, var, num_user);
}

void Vistaplus::AUC_NonSessionTransition(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_NonSessionTransition(AUC_test, var, num_user);
}

void Vistaplus::AUC_coldFull(double* AUC_test, double* var, int* num_user)
{
	project();
	Vista::AUC_coldFull(AUC_test, var, num_user);
}

void Vistaplus::tsne2D(const char* tofilePath)
{
	// get visual factors
	project();

	// get the flat version of visual factors (of items in scope)	
	double* flat_visual_factors = new double [nItems * KK];
	int index = 0;
	for (int i = 0; i < nItems; i ++) {
		for (int k = 0; k < KK; k ++) {
			flat_visual_factors[index ++] = kappa_item[i][k] + kappaV_item[i][k];
		}
	}

	// compress using t-sne
	int num_dim = 2; // fixed to 2-d for visualization
	int N = nItems;
	double perplexity = 30; 
	double theta = 0.5;
	int rand_seed = 1;
    TSNE* tsne = new TSNE();
	fprintf(stderr, "\n  t-SNE: Using random seed: %d\n", rand_seed);
	srand((unsigned int) rand_seed);

	// now fire up the SNE implementation
	double* low_dim_factors = new double [N * num_dim];
	double* costs = new double [N];
	memset(costs, 0, N * sizeof(double));
	tsne->run(flat_visual_factors, N, KK, low_dim_factors, num_dim, perplexity, theta);
	
	// clean up the memory
	delete [] flat_visual_factors;
	delete [] costs;
	delete tsne;

	// save low dimensional factors to file for plot
	FILE* coordF = fopen_(tofilePath, "w");
	for (int item = 0; item < nItems; item ++) {
		string asin = corp->rItemIds[item];
		fprintf(coordF, "%s ", asin.c_str());
		for (int k = 0; k < num_dim; k ++) {
			fprintf(coordF, "%f", low_dim_factors[item * num_dim + k]);
			if (k < num_dim - 1) {
				fprintf(coordF, " ");
			}
		}
		if (item < nItems) {
			fprintf(coordF, "\n");
		}
	}
	fclose(coordF);
	delete [] low_dim_factors;
	fprintf(stderr, "\n  Writen to %s.\n", tofilePath);
}